#include <SPIFFS.h>
#include <FS.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <ArduinoJson.h>
#include <MIDI.h>
#include <WiFi.h>
#include <WiFiClient.h>
#include <LiquidCrystal.h>

MIDI_CREATE_DEFAULT_INSTANCE();
AsyncWebServer server(80);

const byte rs = 33, en = 25, d4 = 32, d5 = 26, d6 = 27, d7 = 14;
const byte in1 = 36, in2 = 39, butt1 = 35, butt2 = 34, outA = 18, outB = 19, outC = 21, led = 4, pi = 23;

LiquidCrystal lcd(rs, en, d4, d5, d6, d7);

//AP mode
bool wifi = false;
const char* ssid = "Defect";
const char* password = "00000000";
/* IP Address details */
IPAddress apIP(192, 6, 6, 6);
IPAddress subnet(255, 255, 255, 0);

int butt1val = 0;
int butt2val = 0;

int presetNr = 0;
int bankNr = 9; // initial bank
int tapMode = 0;

bool tapped = true;
unsigned long previousMillis = 0;
const long interval = 300;

String BankName[25];
String RPName[50];

typedef struct {
  String n;
  byte t;
  byte r;
  byte p;
  byte o;
} preset_type;

preset_type theBanks[125];

void setup() {
  /*LED   */
  pinMode(led, OUTPUT);
  digitalWrite(led, 0);

  /*5V outputs   */
  pinMode(outA, OUTPUT);
  pinMode(outB, OUTPUT);
  pinMode(outC, OUTPUT);

  //  Serial.begin(115200);

  // Route for root / web page
  server.on("/", HTTP_GET, [](AsyncWebServerRequest * request) {

    String inputMessage;
    if (request->hasParam("bn")) {
      BankName[bankNr] = request->getParam("bn")->value();
    }
    if (request->hasParam("b")) {
      inputMessage = request->getParam("b")->value();
      bankNr = inputMessage.toInt();
      presetNr = bankNr * 5;
      loadPreset(presetNr);
    }
    if (request->hasParam("rpi")) {
      inputMessage = request->getParam("rpi")->value();
      int rpIndex = inputMessage.toInt();
      RPName[rpIndex] = request->getParam("rpn")->value();
    }
    if (request->hasParam("n0")) {
      int i = presetNr;
      theBanks[i].n = request->getParam("n0")->value();
      inputMessage = request->getParam("r0")->value();
      theBanks[i].r = inputMessage.toInt();
      inputMessage = request->getParam("p0")->value();
      theBanks[i].p = inputMessage.toInt();
      inputMessage = request->getParam("t0")->value();
      theBanks[i].t = inputMessage.toInt();
      inputMessage = request->getParam("o0")->value();
      theBanks[i].o = inputMessage.toInt();
      i++;
      theBanks[i].n = request->getParam("n1")->value();
      inputMessage = request->getParam("r1")->value();
      theBanks[i].r = inputMessage.toInt();
      inputMessage = request->getParam("p1")->value();
      theBanks[i].p = inputMessage.toInt();
      inputMessage = request->getParam("t1")->value();
      theBanks[i].t = inputMessage.toInt();
      inputMessage = request->getParam("o1")->value();
      theBanks[i].o = inputMessage.toInt();
      i++;
      theBanks[i].n = request->getParam("n2")->value();
      inputMessage = request->getParam("r2")->value();
      theBanks[i].r = inputMessage.toInt();
      inputMessage = request->getParam("p2")->value();
      theBanks[i].p = inputMessage.toInt();
      inputMessage = request->getParam("t2")->value();
      theBanks[i].t = inputMessage.toInt();
      inputMessage = request->getParam("o2")->value();
      theBanks[i].o = inputMessage.toInt();
      inputMessage = request->getParam("r2")->value();
      theBanks[i].r = inputMessage.toInt();
      inputMessage = request->getParam("p2")->value();
      theBanks[i].p = inputMessage.toInt();
      inputMessage = request->getParam("t2")->value();
      theBanks[i].t = inputMessage.toInt();
      inputMessage = request->getParam("o2")->value();
      theBanks[i].o = inputMessage.toInt();
      i++;
      theBanks[i].n = request->getParam("n3")->value();
      inputMessage = request->getParam("r3")->value();
      theBanks[i].r = inputMessage.toInt();
      inputMessage = request->getParam("p3")->value();
      theBanks[i].p = inputMessage.toInt();
      inputMessage = request->getParam("t3")->value();
      theBanks[i].t = inputMessage.toInt();
      inputMessage = request->getParam("o3")->value();
      theBanks[i].o = inputMessage.toInt();
      i++;
      theBanks[i].n = request->getParam("n4")->value();
      inputMessage = request->getParam("r4")->value();
      theBanks[i].r = inputMessage.toInt();
      inputMessage = request->getParam("p4")->value();
      theBanks[i].p = inputMessage.toInt();
      inputMessage = request->getParam("t4")->value();
      theBanks[i].t = inputMessage.toInt();
      inputMessage = request->getParam("o4")->value();
      theBanks[i].o = inputMessage.toInt();
    }

    request->send(SPIFFS, "/index.html", String(), false, processor);
  });
  // Route to load image
  server.on("/d.png", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(SPIFFS, "/d.png", "image/png");
  });
  server.on("/favicon.ico", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(SPIFFS, "/favicon.png", "image/png");
  });
  server.on("/favicon.png", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(SPIFFS, "/favicon.png", "image/png");
  });
  server.on("/style.css", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(SPIFFS, "/style.css", "text/css");
  });
  server.on("/data.txt", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(SPIFFS, "/data.txt", "text/plain");
  });
  server.on("/save", HTTP_GET, [](AsyncWebServerRequest * request) {
    writeData();
    request->send(200, "text/plain", "Saved!");
  });

  // Start services
  SPIFFS.begin();
  //server.begin();
  lcd.begin(16, 2);
  MIDI.begin(1);

  //read data
  readData();

  //initial display
  lcd.clear();
  lcd.print(BankName[bankNr]);
  lcd.setCursor(0, 1);                //lcd.setCursor(col, row)
  lcd.print(bankNr + 1);
}

void loop() {

  unsigned long currentMillis = millis();
  if (currentMillis - previousMillis >= interval) {
    previousMillis = currentMillis;
    butt1val = analogRead(butt1);
    butt2val = analogRead(butt2);

    if (butt1val < 4000) {
      if (butt1val < 100) {                                 //A
        presetNr = bankNr * 5 + 0;
        loadPreset(presetNr);

      } else if ((butt1val > 450) && (butt1val < 650)) {    //B
        presetNr = bankNr * 5 + 1;
        loadPreset(presetNr);

      } else if ((butt1val > 1750) && (butt1val < 2000)) {  //C
        presetNr = bankNr * 5 + 2;
        loadPreset(presetNr);

      } else if ((butt1val > 2900) && (butt1val < 3100)) {  //D
        presetNr = bankNr * 5 + 3;
        loadPreset(presetNr);

      }
    }
    if (butt2val < 4000) {
      if (butt2val < 100) {                                 //E
        presetNr = bankNr * 5 + 4;
        loadPreset(presetNr);

      } else if ((butt2val > 450) && (butt2val < 650)) {    //tap
        tapFunc();

      } else if ((butt2val > 1750) && (butt2val < 2000)) {  //b-
        if (bankNr > 0) {
          bankNr = bankNr - 1;
        } else {
          bankNr = 24;
        }

        lcd.clear();
        lcd.print(BankName[bankNr]);
        lcd.setCursor(0, 1);                //lcd.setCursor(col, row)
        lcd.print(bankNr + 1);

      } else if ((butt2val > 2900) && (butt2val < 3100)) {  //b+
        if (bankNr < 24) {
          bankNr = bankNr + 1;
        } else {
          bankNr = 0;
        }

        lcd.clear();
        lcd.print(BankName[bankNr]);
        lcd.setCursor(0, 1);                //lcd.setCursor(col, row)
        lcd.print(bankNr + 1);
      }
    }

    bool zero = digitalRead(0);

    if (!zero) {
      if (wifi == false) {
        //turn wifi on
        WiFi.mode(WIFI_AP);
        WiFi.softAP(ssid, password);
        digitalWrite(led, 1);
        delay(100);
        WiFi.softAPConfig(apIP, apIP, subnet);
        delay(100);
        wifi = true;
        server.begin();
      } else {
        //turn wifi off
        WiFi.mode(WIFI_OFF);
        delay(1);
        digitalWrite(led, 0);
        wifi = false;
      }
    }

  }
  delay(1);
}

// Replaces placeholders in web page
String processor(const String & var) {
  if (var == "BANKNAME") {
    return BankName[bankNr];
  }
  if (var == "RAM") {
    uint32_t Rfree = ESP.getFreeHeap();
    uint32_t Rtot = ESP.getHeapSize();
    String ram = String(Rfree);
    ram += " of ";
    ram += String(Rtot);
    return ram;
  }
  if (var == "BANKLIST") {
    String buff = "";
    for (int i = 0; i <= 24; i++) {
      buff += "<a href=\"?b=" + String(i) + "\"><li>" + BankName[i] + "</li></a>";
    }
    return buff;
  }
  if (var == "RPLIST") {
    String buff1 = "";
    for (int i = 0; i <= 49; i++) {
      buff1 += "<li id=\"" + String(i) + "\"  onclick=\"sel(this.id, this.innerText)\">" + RPName[i] + "</li>";
    }
    return buff1;
  }
  if (var == "TABLE") {
    String buf = "";
    buf += "<tr class=\"k\">";
    buf += "<td></td>";
    buf += "<td><h2>A</h2></td>";
    buf += "<td><h2>B</h2></td>";
    buf += "<td><h2>C</h2></td>";
    buf += "<td><h2>D</h2></td>";
    buf += "<td><h2>E</h2></td>";
    buf += "</tr>";
    buf += "<tr>";
    buf += "<td class=\"k\">Name:</td>";
    for (int i = 0; i <= 4; i++) {
      buf += "<td><input type=\"text\" name=\"n" + String(i) + "\" value=\"" + theBanks[(presetNr + i)].n + "\" maxlength=8></td>";
    }
    buf += "</tr>";
    buf += "<tr>";
    buf += "<td class=\"k\">RP155:</td>";
    for (int i = 0; i <= 4; i++) {
      buf += "<td><input type=\"number\" name=\"r" + String(i) + "\" min=\"1\" max=\"99\"  value=\"" + String(theBanks[(presetNr + i)].r) + "\"></td>";
    }
    buf += "</tr>";
    buf += "<tr>";
    buf += "<td class=\"k\">RP80:</td>";
    for (int i = 0; i <= 4; i++) {
      buf += "<td><input type=\"number\" name=\"p" + String(i) + "\" min=\"1\" max=\"50\"  value=\"" + String(theBanks[(presetNr + i)].p) + "\"></td>";
    }
    buf += "</tr>";
    buf += "<tr>";
    buf += "<td class=\"k\">Tap:</td>";
    for (int i = 0; i <= 4; i++) {
      buf += "<td><input type=\"number\" name=\"t" + String(i) + "\" min=\"1\" max=\"10\"  value=\"" + String(theBanks[(presetNr + i)].t) + "\"></td>";
    }
    buf += "</tr>";
    buf += "<tr>";
    buf += "<td class=\"k\">Out:</td>";
    for (int i = 0; i <= 4; i++) {
      buf += "<td><input type=\"number\" name=\"o" + String(i) + "\" min=\"0\" max=\"255\"  value=\"" + String(theBanks[(presetNr + i)].o) + "\"></td>";
    }
    buf += "</tr>";
    return buf;
  }
  return String();
}

void readData() {

  File fl = SPIFFS.open("/data.txt", "r");
  DynamicJsonDocument doc(16200);
  deserializeJson(doc, fl);

  for (int i = 0; i <= 24; i++) {
    BankName[i] = doc["banks"][i].as<String>();
  }
  for (int i = 0; i <= 49; i++) {
    RPName[i] = doc["rp"][i].as<String>();
  }
  for (int i = 0; i <= 124; i++) {
    theBanks[i].n = doc["presets"][i]["n"].as<String>();
    theBanks[i].t = doc["presets"][i]["t"];
    theBanks[i].r = doc["presets"][i]["r"];
    theBanks[i].p = doc["presets"][i]["p"];
    theBanks[i].o = doc["presets"][i]["o"];
  }

  fl.close();
}

void writeData() {

  File fl = SPIFFS.open("/data.txt", "w");
  DynamicJsonDocument doc(16200);

  JsonArray presets = doc.createNestedArray("presets");
  for (int i = 0; i <= 124; i++) {
    JsonObject apreset = presets.createNestedObject();
    apreset["n"] = theBanks[i].n;
    apreset["t"] = theBanks[i].t;
    apreset["r"] = theBanks[i].r;
    apreset["p"] = theBanks[i].p;
    apreset["o"] = theBanks[i].o;
  }

  JsonArray banks = doc.createNestedArray("banks");
  for (int i = 0; i <= 24; i++) {
    banks.add(BankName[i]);
  }

  JsonArray rp = doc.createNestedArray("rp");
  for (int i = 0; i <= 49; i++) {
    rp.add(RPName[i]);
  }

  serializeJson(doc, fl);
  fl.close();
}

void loadPreset(byte p) {
  lcd.clear();

  byte rp155Preset = theBanks[p].r - 1;
  byte rp80Preset = theBanks[p].p;
  MIDI.sendProgramChange(p, 2);
  MIDI.sendProgramChange(p, 3);
  MIDI.sendProgramChange(rp155Preset, 1);
  MIDI.sendProgramChange(rp80Preset, 4);
  tapMode = theBanks[p].t;
  bool outA = theBanks[p].o & 1;
  bool outB = theBanks[p].o & (1 << 1);
  bool outC = theBanks[p].o & (1 << 2);
  digitalWrite(outA, outA);
  digitalWrite(outB, outB);
  digitalWrite(outC, outC);
  bool PiReady = true;
  PiReady = digitalRead(pi);

  lcd.setCursor(0, 0);
  lcd.print(BankName[bankNr]);
  lcd.setCursor(9, 0);
  switch (tapMode) {
    case 3:
      // "TAP "
      lcd.print("tap");
      break;
    case 1:
      // "reverb"
      lcd.print("reverb");
      break;
    case 5:
      // "MUTE"
      lcd.print("mute");
    case 6:
      // "M V3"
      lcd.print("mute V3");
      break;
    case 7:
      // "M V2"
      lcd.print("mute V2");
      break;
    case 2:
      // "verb + RP"
      lcd.print("verb&RP");
      break;
    case 4:
      // "RP bypass"
      lcd.print("RP off");
      break;
  }
  lcd.setCursor(0, 1);
  lcd.print(theBanks[p].n);
  lcd.setCursor(9, 1);
  if (!PiReady) {
    lcd.print(RPName[rp155Preset]);
  } else {
    lcd.print("-wait-");
  }
}

void tapFunc() {
  byte pcRP = theBanks[presetNr].r - 1;

  switch (tapMode) {
    case 3:
      // "TAP "
      MIDI.sendControlChange(64, 100, 2);
      MIDI.sendControlChange(64, 100, 3);
      break;
    case 1:
      // "reverb"
      if (tapped) {
        MIDI.sendControlChange(22, 127, 2);
        MIDI.sendControlChange(22, 127, 3);
      } else {
        MIDI.sendControlChange(22, 0, 2);
        MIDI.sendControlChange(22, 0, 3);
      }
      break;
    case 5:
      // "MUTE"
      if (tapped) {
        MIDI.sendProgramChange(127, 2);    //V-amp 127 = tuner
        MIDI.sendProgramChange(127, 3);
      } else {
        MIDI.sendProgramChange(presetNr, 2);
        MIDI.sendProgramChange(presetNr, 3);
      }
      break;
    case 6:
      // "M V3"
      if (tapped) {
        MIDI.sendProgramChange(127, 3);
      } else {
        MIDI.sendProgramChange(presetNr, 3);
      }
      break;
    case 7:
      // "M V2"
      if (tapped) {
        MIDI.sendProgramChange(127, 2);
      } else {
        MIDI.sendProgramChange(presetNr, 2);
      }
      break;
    case 2:
      // "verb + RP"
      if (tapped) {
        MIDI.sendControlChange(22, 127, 2);
        MIDI.sendControlChange(22, 127, 3);
        MIDI.sendProgramChange(pcRP, 1);
      } else {
        MIDI.sendControlChange(22, 0, 2);
        MIDI.sendControlChange(22, 0, 3);
        MIDI.sendProgramChange(10, 1);
      }
      break;
    case 4:
      // "RP bypass"
      if (tapped) {
        MIDI.sendProgramChange(10, 1);    //RP10 = bypass
      } else {
        MIDI.sendProgramChange(pcRP, 1);
      }
      break;
  }
  tapped = !tapped;
}
